@@ -1,8 +1,32 @@
-/* -*- mode: c; c-basic-offset: 8 -*- */
+/* -*- Mode: C; c-basic-offset: 8; indent-tabs-mode: nil -*- */
+/* vim: set filetype=c tabstop=8 shiftwidth=4 expandtab: */
+
 
 #ifndef ECL_ECL_INL_H
 #define ECL_ECL_INL_H
 
+#ifdef ECL_COMPLEX_FLOAT
+/* We need this include for the constant I. */
+#include <complex.h>
+#endif
+
+/*
+ * If minimum unnormalized floating point values defined in ISO C11
+ * are not declared, we use the normalized ones as the next best
+ * portable approximation.
+ */
+#include <float.h>
+
+#ifndef FLT_TRUE_MIN
+# define FLT_TRUE_MIN FLT_MIN
+#endif
+#ifndef DBL_TRUE_MIN
+# define DBL_TRUE_MIN DBL_MIN
+#endif
+#ifndef LDBL_TRUE_MIN
+# define LDBL_TRUE_MIN LDBL_MIN
+#endif
+
 /*
  * Loops over a proper list. Complains on circularity
  */
@@ -73,11 +97,11 @@
                 cl_fixnum padding[(sizeof(struct ecl_base_string)+3)/4*4]; \
         } name[]
 
-#define ecl_def_string_array_elt(chars) {                      \
+#define ecl_def_string_array_elt(chars) { {                      \
                 (int8_t)t_base_string, 0, ecl_aet_bc, 0,            \
                         ECL_NIL, (cl_index)(sizeof(chars))-1,      \
                         (cl_index)(sizeof(chars))-1,            \
-                        (ecl_base_char*)(chars) }
+                        (ecl_base_char*)(chars) } }
 
 #define ecl_def_ct_base_string(name,chars,len,static,const)     \
         static const struct ecl_base_string name ## _data = {    \
@@ -98,28 +122,48 @@
                 (double)(f) };                                  \
         static const cl_object name = (cl_object)(& name ## _data)
 
-#define ecl_def_ct_long_float(name,f,static,const)			\
-        static const struct ecl_long_float name ## _data = {		\
-                (int8_t)t_longfloat, 0, 0, 0,				\
-                (long double)(f) };					\
+#define ecl_def_ct_long_float(name,f,static,const)                      \
+        static const struct ecl_long_float name ## _data = {            \
+                (int8_t)t_longfloat, 0, 0, 0,                           \
+                (long double)(f) };                                     \
         static const cl_object name = (cl_object)(& name ## _data)
 
-#define ecl_def_ct_ratio(name,num,den,static,const)			\
-        static const struct ecl_ratio name ## _data = {			\
-                (int8_t)t_ratio, 0, 0, 0,				\
-                den, num };						\
+#define ecl_def_ct_ratio(name,num,den,static,const)                     \
+        static const struct ecl_ratio name ## _data = {                 \
+                (int8_t)t_ratio, 0, 0, 0,                               \
+                den, num };                                             \
         static const cl_object name = (cl_object)(& name ## _data)
 
-#define ecl_def_ct_complex(name,real,imag,static,const)			\
-        static const struct ecl_complex name ## _data = {		\
-                (int8_t)t_complex, 0, 0, 0,				\
-                (cl_object)real, (cl_object)imag };			\
+#define ecl_def_ct_complex(name,real,imag,static,const)                 \
+        static const struct ecl_complex name ## _data = {               \
+                (int8_t)t_complex, 0, 0, 0,                             \
+                (cl_object)real, (cl_object)imag };                     \
         static const cl_object name = (cl_object)(& name ## _data)
 
+#ifdef ECL_COMPLEX_FLOAT
+#define ecl_def_ct_csfloat(name,f,static,const)                         \
+        static const struct ecl_csfloat name ## _data = {               \
+                (int8_t)t_csfloat, 0, 0, 0,                             \
+                (float _Complex)(f) };                                  \
+        static const cl_object name = (cl_object)(& name ## _data)
+
+#define ecl_def_ct_cdfloat(name,f,static,const)                         \
+        static const struct ecl_cdfloat name ## _data = {               \
+                (int8_t)t_cdfloat, 0, 0, 0,                             \
+                (double _Complex)(f) };                                 \
+        static const cl_object name = (cl_object)(& name ## _data)
+
+#define ecl_def_ct_clfloat(name,f,static,const)                         \
+        static const struct ecl_clfloat name ## _data = {               \
+                (int8_t)t_clfloat, 0, 0, 0,                             \
+                (long double _Complex)(f) };                            \
+        static const cl_object name = (cl_object)(& name ## _data)
+#endif
+
 #define ecl_def_ct_vector(name,type,raw,len,static,const)               \
-        static const struct ecl_vector name ## _data = {                 \
+        static const struct ecl_vector name ## _data = {                \
                 (int8_t)t_vector, 0, (type), 0,                         \
-                ECL_NIL, (cl_index)(len), (cl_index)(len),                 \
+                ECL_NIL, (cl_index)(len), (cl_index)(len),              \
                 ecl_cast_ptr(cl_object*,raw), 0 };                      \
         static const cl_object name = (cl_object)(& name ## _data)
 
